import asyncio
import aiohttp
import json
from .utils import to_instruction, to_json_desc, construct_request_messages

def generate_request_data(get_settings, request_runtime_ctx):
    #default request data
    request_data = {
        "url": "https://api.openai.com/v1/chat/completions",
        "headers": {
            "Authorization": None,
            "Content-Type": "application/json",
            "Accept": "text/event-stream",
        },
        "data": {
            "model": "gpt-3.5-turbo",
            "temperature": 1,
            "stream": True
        },
        "proxy": None
    }
    #model settings
    model_settings_data = get_settings("model_settings")
    if "auth" in model_settings_data and "api_key" in model_settings_data["auth"]:
        request_data["headers"]["Authorization"] = f"Bearer { model_settings_data['auth']['api_key'] }"
    else:
        raise Exception("[GPT Request] Missing 'model_settings.auth.api_key' in request runtime_ctx.")
    if "url" in model_settings_data and isinstance(model_settings_data["url"], str):
        request_data["url"] = model_settings_data["url"]
    if "options" in model_settings_data and isinstance(model_settings_data["options"], dict):
        request_data["data"].update(model_settings_data["options"])
    #proxy
    request_data["proxy"] = get_settings("proxy")
    #construct request messages
    request_messages = construct_request_messages(request_runtime_ctx)
    #transform request messages to model required format
        #nah, no need for OpenAI GPT since it is the base format
    request_data["data"].update({ "messages": request_messages })
    return request_data

async def request_model(request_data):
    async with aiohttp.ClientSession() as session:
        async with session.post(
            request_data["url"],
            headers = request_data["headers"],
            json = request_data["data"],
            proxy = request_data["proxy"],
        ) as response:
            async for chunk in response.content.iter_chunks():
                items = chunk[0].decode("utf-8").split("\n\n")
                items = [item for item in items if item != '']
                for item in items:
                    if item[:6] == 'data: ':
                        delta = item[6:]
                        if delta == '[DONE]':
                            yield { "is_done": True }
                            break
                        else:
                            yield { "is_done": False, "delta" : json.loads(delta) }
                    else:
                        raise Exception(f"[GPT Request] Error occurred when request API:\n { str(item) }")
                
def content_mapping(event_type: str, origin_data: dict):
    if event_type == "delta":
        if "delta" in origin_data["choices"][0] and "content" in origin_data["choices"][0]["delta"]:
            return origin_data["choices"][0]["delta"]["content"]
        else:
            return None
    elif event_type == "done":
        return origin_data["choices"][0]["content"]

async def broadcast_response(response_generater):
    response_message = {}
    full_response_message = {}
    async for data in response_generater:
        if not data["is_done"]:
            if "choices" not in data["delta"]:
                raise Exception(f"[GPT Request] Error occurred when request API:\n { str(data) }")
            else:
                delta_content = data["delta"]["choices"][0]
                if "delta" in delta_content:
                    full_response_message.update(data["delta"])
                    for key, value in delta_content["delta"].items():
                        value = response_message[key] + value if key in response_message else value
                        response_message.update({ key: value })
                    yield { "event": "response:delta_origin", "data": data["delta"] }
                    response_content = content_mapping("delta", data["delta"])
                    if response_content != None and response_content != '':
                        yield { "event": "response:delta", "data": response_content }
        else:
            del full_response_message["choices"][0]["delta"]
            full_response_message["choices"][0].update(response_message)
            yield { "event": "response:done_origin", "data": full_response_message }
            yield { "event": "response:done", "data": content_mapping("done", full_response_message) }

def export():
    return ('gpt', {
        "generate_request_data": generate_request_data,
        "request_model": request_model,
        "broadcast_response": broadcast_response,
    })


import os
import asyncio
import threading
import queue
import json
from configparser import ConfigParser

from ..utils import RuntimeCtx, RuntimeCtxNamespace, PluginManager, AliasManager, to_json_desc, find_json
from .._global import global_plugin_manager

class Request(object):
    def __init__(
            self,
            *,
            parent_plugin_manager: object = global_plugin_manager,
            parent_request_runtime_ctx:object = None
        ):
        self.request_runtime_ctx = RuntimeCtx(parent = parent_request_runtime_ctx)
        self.response_cache = {
            "prompt": {},
            "type": None,
            "reply": None,
        }
        # Plugin Manager
        self.plugin_manager = PluginManager(parent = parent_plugin_manager)
        # Namespace
        self.model_settings = RuntimeCtxNamespace("$.settings.model_settings", self.plugin_manager.plugins_runtime_ctx, return_to = self)
        self.system = RuntimeCtxNamespace("system", self.request_runtime_ctx, return_to = self)
        self.headline = RuntimeCtxNamespace("headline", self.request_runtime_ctx, return_to = self)
        self.chat_history = RuntimeCtxNamespace("chat_history", self.request_runtime_ctx, return_to = self)
        self.prompt_input = RuntimeCtxNamespace("prompt_input", self.request_runtime_ctx, return_to = self)
        self.prompt_information = RuntimeCtxNamespace("prompt_information", self.request_runtime_ctx, return_to = self)
        self.prompt_instruction = RuntimeCtxNamespace("prompt_instruction", self.request_runtime_ctx, return_to = self)
        self.prompt_output = RuntimeCtxNamespace("prompt_output", self.request_runtime_ctx, return_to = self)
        # Alias
        self.alias_manager = AliasManager(self)
        self._register_default_alias(self.alias_manager)

    def _register_default_alias(self, alias_manager):
        alias_manager.register("set_model", lambda key, value: self.plugin_manager.set_settings(f"model_settings.{ key }", value))
        alias_manager.register("set_model_name", lambda model_name: self.plugin_manager.set_settings("model_settings.model_name", model_name))
        alias_manager.register("set_model_auth", lambda model_auth: self.plugin_manager.set_settings("model_settings.auth", model_auth))
        alias_manager.register("set_model_url", lambda model_url: self.plugin_manager.set_settings("model_settings.url", model_url))
        alias_manager.register("set_model_options", lambda model_options: self.plugin_manager.set_settings("model_settings.options", model_options))
        alias_manager.register("set_system", self.system.assign)
        alias_manager.register("set_headline", self.headline.assign)
        alias_manager.register("set_chat_history", self.chat_history.assign)
        alias_manager.register("input", self.prompt_input.assign)
        alias_manager.register("info", self.prompt_information.assign)
        alias_manager.register("instruct", self.prompt_instruction.assign)
        alias_manager.register("output", self.prompt_output.assign)
        alias_manager.register("set_proxy", lambda proxy_setting: self.plugin_manager.set_settings("proxy", proxy_setting))
        
    def get_event_generator(self):
        # Erase response cache
        self.response_cache = {
            "prompt": {},
            "type": None,
            "reply": None,
        }
        # Confirm model name
        model_name = self.plugin_manager.get_settings("model_settings.model_name")
        if not model_name:
            raise Exception(f"[Request] 'model_name' must be set into settings_runtime_ctx 'model_settings'.")
        model_name = model_name.lower()
        # Load request plugin by model name
        request_plugin = self.plugin_manager.get("request", model_name)
        # Generate request data
        request_data = request_plugin["generate_request_data"](
            request_runtime_ctx = self.request_runtime_ctx,
            get_settings = self.plugin_manager.get_settings,
        )
        if self.plugin_manager.get_settings("is_debug") == True:
            print("[Request Data]\n", json.dumps(request_data["data"], indent=4, ensure_ascii=False))
        # Cache simple prompt
        self.response_cache["prompt"]["input"] = self.prompt_input.get()
        self.response_cache["prompt"]["output"] = self.prompt_output.get()
        # Request and get response generator
        response_generator = request_plugin["request_model"](request_data)
        # Broadcast response
        broadcast_event_generator = request_plugin["broadcast_response"](response_generator)
        self.response_cache["type"] = self.request_runtime_ctx.get("response:type")
        # Reset request runtime_ctx
        self.request_runtime_ctx.empty()
        return broadcast_event_generator

    async def get_result_async(self):
        is_debug = self.plugin_manager.get_settings("is_debug")
        event_generator = self.get_event_generator()
        if is_debug:
            print("[Realtime Response]\n")
        async for response in event_generator:
            if response["event"] == "response:delta" and is_debug:
                print(response["data"], end="")
            if response["event"] == "response:done":
                if is_debug:
                    print("\n--------------------------\n")
                    print("[Final Response]\n", response["data"], "\n--------------------------\n")
                self.response_cache["reply"] = response["data"]
        if self.response_cache["type"] == "JSON":
            try:
                self.response_cache["reply"] = json.loads(find_json(self.response_cache["reply"]))
            except json.JSONDecodeError as e:
                try:
                    fixed_result = await self.start\
                        .input({
                            "target": self.response_cache["prompt"]["input"],
                            "format": to_json_desc(self.response_cache["prompt"]["output"]),
                            "origin JSON String": self.response_cache["reply"],
                            "error": e.msg,
                            "position": e.pos,
                        })\
                        .output('Fixed JSON String can be parsed by Python only without explanation and decoration.')\
                        .get_result_async()
                    fixed_result = json.loads(find_json(fixed_result))
                    return fixed_result
                except Exception as e:
                    raise Exception(f"[Agent Request] Error still occured when try to fix JSON decode error: { str(e) }")
        return self.response_cache["reply"]

    def get_result(self):
        reply_queue = queue.Queue()
        def start_in_theard():
            asyncio.set_event_loop(asyncio.new_event_loop())
            reply = asyncio.get_event_loop().run_until_complete(self.get_result_async())
            reply_queue.put_nowait(reply)
        theard = threading.Thread(target=start_in_theard)
        theard.start()
        theard.join()        
        reply = reply_queue.get_nowait()
        return reply

    def start(self):
        return self.get_result()


from .utils import componentABC

class Status(componentABC):
    def __init__(self, agent):
        pass

    def export(self):
        return {
            "prefix": None,
            "suffix": None,
            "alias": {},
        }

def export():
    return ("Status", Status)


import datetime
import json
import asyncio
import threading
import queue
from ..Request import Request
from ..WebSocket import WebSocketServer
from ..utils import RuntimeCtx, StorageDelegate, PluginManager, AliasManager, IdGenerator, to_json_desc, find_json, check_version

class Agent(object):
    def __init__(
        self,
        *,
        agent_id: str=None,
        auto_save: bool=False,
        parent_agent_runtime_ctx: object,
        global_storage: object,
        global_websocket_server: object,
        parent_plugin_manager: object,
    ):
        # Integrate
        self.global_storage = global_storage
        self.global_websocket_server = global_websocket_server
        self.plugin_manager = PluginManager(parent = parent_plugin_manager)
        self.alias_manager = AliasManager(self)
        self.agent_runtime_ctx = RuntimeCtx(parent = parent_agent_runtime_ctx)
        self.request_runtime_ctx = RuntimeCtx()
        self.request = Request(
            parent_plugin_manager = self.plugin_manager,
            parent_request_runtime_ctx = self.request_runtime_ctx,
        )
        # Agent Id
        if agent_id == None:
            self.agent_id = IdGenerator("agent").create()
        else:
            self.agent_id = agent_id
        # Agent Storage
        self.agent_storage = StorageDelegate(
            db_name = self.agent_id,
            plugin_manager = self.plugin_manager,
        )
        # Load Saved agent_runtime_ctx
        self.agent_runtime_ctx.update_by_dict(self.agent_storage.table("agent_runtime_ctx").get())
        # Set Agent Auto Save Setting
        self.agent_runtime_ctx.set("agent_auto_save", auto_save)
        # Version Check In Debug Model
        if self.plugin_manager.get_settings("is_debug"):
            check_version_record = self.global_storage.get("agently", "check_version_record")
            today = str(datetime.date.today())
            if check_version_record != today:
                check_version(self.global_storage, today)
        # Agent Request Prefix & Suffix
        self.agent_request_prefix = []
        self.agent_request_suffix = []
        # Register Default Request Alias to Agent
        self.request._register_default_alias(self.alias_manager)
        # Install Agent Components
        self.refresh_plugins()

    def refresh_plugins(self):
        # Agent Components
        agent_components = self.plugin_manager.get("agent_component")
        component_toggles = self.plugin_manager.get_settings("component_toggles")
        for agent_component_name, AgentComponentClass in agent_components.items():
            # Skip component_toggles Those Be Toggled Off
            if agent_component_name in component_toggles and component_toggles[agent_component_name] == False:
                    continue
            # Attach Component
            agent_component_instance = AgentComponentClass(agent = self)
            setattr(self, agent_component_name, agent_component_instance)
            component_export = agent_component_instance.export()
            # Register export_prefix, export_suffix
            if component_export["prefix"]:
                if isinstance(component_export["prefix"], list):
                    self.agent_request_prefix.extend(component_export["prefix"])
                elif callable(component_export["prefix"]):
                    self.agent_request_prefix.append(component_export["prefix"])
            if component_export["suffix"]:
                if isinstance(component_export["suffix"], list):
                    self.agent_request_suffix.extend(component_export["suffix"])
                elif callable(component_export["suffix"]):
                    self.agent_request_suffix.append(component_export["suffix"])
            # Register Alias
            if component_export["alias"]:
                for alias_name, alias_info in component_export["alias"].items():
                    self.alias_manager.register(
                        alias_name,
                        alias_info["func"],
                        return_value = alias_info["return_value"] if "return_value" in alias_info else False,
                    )

    def toggle_auto_save(self, is_enabled: bool):
        self.agent_runtime_ctx.set("agent_auto_save", is_enabled)
        return self

    def save(self):
        self.agent_storage.table("agent_runtime_ctx").update_by_dict(self.agent_runtime_ctx.get()).save()
        return self

    def set_settings(self, settings_key: str, settings_value: any):
        self.plugin_manager.set_settings(settings_key, settings_value)
        return self

    async def start_async(self):
        is_debug = self.plugin_manager.get_settings("is_debug")
        # Auto Save Agent runtime_ctx
        if self.agent_runtime_ctx.get("agent_auto_save") ==  True:
            self.save()
        # Call Prefix Funcs to Prepare Prefix Data(From agent_runtime_ctx To request_runtime_ctx)
        for prefix_func in self.agent_request_prefix:
            prefix_data = await prefix_func() if asyncio.iscoroutinefunction(prefix_func) else prefix_func()
            if prefix_data != None:
                if isinstance(prefix_data, tuple) and isinstance(prefix_data[0], str) and prefix_data[1] != None:
                    self.request.request_runtime_ctx.update(prefix_data[0], prefix_data[1])
                else:
                    raise Exception("[Agent Component] Prefix return data error: only accept None or Tuple('<Request Namespace>', <Update Dict>)")

        # Request
        event_generator = self.request.get_event_generator()
    
        # Call Suffix Func to Handle Response Events
        if is_debug:
            print("[Realtime Response]\n")
        async def call_request_suffix(response):
            for suffix_func in self.agent_request_suffix:
                if asyncio.iscoroutinefunction(suffix_func):
                    await suffix_func(response["event"], response["data"]) 
                else:
                    suffix_func(response["event"], response["data"])

        async for response in event_generator:
            if response["event"] == "response:delta" and is_debug:
                print(response["data"], end="")
            await call_request_suffix(response)
            if response["event"] == "response:done":
                if self.request.response_cache["reply"] == None:
                    self.request.response_cache["reply"] = response["data"]
                if is_debug:
                    print("\n--------------------------\n")
                    print("[Final Reply]\n", self.request.response_cache["reply"], "\n--------------------------\n")

        await call_request_suffix({ "event": "response:finally", "data": self.request.response_cache })

        # Fix JSON if Required
        if self.request.response_cache["type"] == "JSON":
            try:
                self.request.response_cache["reply"] = json.loads(find_json(self.request.response_cache["reply"]))
                if is_debug:
                    print("[Parse JSON to Dict] Done")
                    print("\n--------------------------\n")
            except json.JSONDecodeError as e:
                try:
                    fixed_result = await self.request\
                        .input({
                            "target": self.request.response_cache["prompt"]["input"],
                            "format": to_json_desc(self.request.response_cache["prompt"]["output"]),
                            "origin JSON String": self.request.response_cache["reply"] ,
                            "error": e.msg,
                            "position": e.pos,
                        })\
                        .output('Fixed JSON String can be parsed by Python only without explanation and decoration.')\
                        .get_result_async()
                    self.request.response_cache["reply"] = json.loads(find_json(fixed_result))
                    if is_debug:
                        print("[Parse JSON to Dict] Done")
                        print("\n--------------------------\n")
                except Exception as e:
                    raise Exception(f"[Agent Request] Error still occured when try to fix JSON decode error: { str(e) }")

        self.request_runtime_ctx.empty()
        return self.request.response_cache["reply"]

    def start(self):
        reply_queue = queue.Queue()
        def start_in_theard():
            asyncio.set_event_loop(asyncio.new_event_loop())
            reply = asyncio.get_event_loop().run_until_complete(self.start_async())
            reply_queue.put_nowait(reply)
        theard = threading.Thread(target=start_in_theard)
        theard.start()
        theard.join()        
        reply = reply_queue.get_nowait()
        return reply

    def start_websocket_server(self):
        is_debug = self.plugin_manager.get_settings("is_debug")

        def alias_handler(data: any, response: callable):
            try:
                if isinstance(data["params"], dict):
                    getattr(self, data["name"])(**data["params"])
                elif isinstance(data["params"], (list, tuple, set)):
                    getattr(self, data["name"])(*data["params"])
                else:
                    getattr(self, data["name"])(data["params"])
                response("alias_done")
            except Exception as e:
                if is_debug:
                    print("[Agent WebSocket Server] Error: ", str(e))
        self.global_websocket_server.add_event_handler(self.agent_id, "alias", alias_handler)

        def start_handler(data: any, response: callable):
            def write_message(event: str, data: any):
                try:
                    response(event, data)
                except Exception as e:
                    if is_debug:
                        print("[Agent WebSocket Server] Error", str(e))
            self\
                .on_delta(
                    lambda data: write_message("delta", data)
                )\
                .on_done(
                    lambda data: write_message("done", data)
                )\
                .start()
        self.global_websocket_server.add_event_handler(self.agent_id, "start", start_handler)

        if is_debug:
            print(f"[WebSocket Server] Event listeners of agent '{ self.agent_id }' are on.")

        if self.global_websocket_server.status == 0:
            if is_debug:
                print(f"[WebSocket Server] WebSocket server started at { self.global_websocket_server.host }:{ self.global_websocket_server.port }.")
            self.global_websocket_server.start()            

    def stop_websocket_server(self):
        self.global_websocket_server.remove_event_handler(self.agent_id)
        if self.plugin_manager.get_settings("is_debug"):
            print(f"[WebSocket Server] Agent '{ self.agent_id } websocket server stoped.'")


from configparser import ConfigParser

from ..utils import PluginManager, RuntimeCtx
from .._global import global_plugin_manager, global_storage, global_websocket_server
from .Agent import Agent

class AgentFactory(object):
    def __init__(self, *, parent_plugin_manager:object = global_plugin_manager, is_debug = False):
        #runtime ctx
        self.factory_agent_runtime_ctx = RuntimeCtx()
        self.settings = RuntimeCtx()
        self.key_chain = RuntimeCtx()

        #use plugin manager
        self.plugin_manager = PluginManager(parent = parent_plugin_manager)

        #use global storage
        self.global_storage = global_storage

        #use global websocket server
        self.global_websocket_server = global_websocket_server

        #debug
        self.set_settings("is_debug", is_debug)

    def create_agent(self, agent_id = None):
        return Agent(
            agent_id = agent_id,
            parent_agent_runtime_ctx = self.factory_agent_runtime_ctx,
            global_storage = self.global_storage,
            global_websocket_server = self.global_websocket_server,
            parent_plugin_manager = self.plugin_manager
        )

    def register_plugin(self, module_name: str, plugin_name: str, plugin: callable):
        self.plugin_manager.register(module_name, plugin_name, plugin)
        return self

    def set_settings(self, settings_key: str, settings_value: any):
        self.plugin_manager.set_settings(settings_key, settings_value)
        return self

    def toggle_component(self, component_name, is_enabled):
        self.plugin_manager.set_settings(f"component_toggles.{ component_name }", is_enabled)
        return self

    def set_proxy(self, proxy_setting: any):
        self.plugin_manager.set_settings("proxy", proxy_setting)
        return self


import uuid
from .utils import componentABC

class Session(componentABC):
    def __init__(self, agent):
        self.agent = agent
        self.current_session_id = None
        self.full_chat_history = []
        self.recent_chat_history = []

    def toggle_auto_save(self, is_enabled: bool):
        self.agent.plugin_manager.set_settings("component_settings.Session.auto_save", is_enabled)
        return self.agent

    def set_max_length(self, max_length: int):
        self.agent.plugin_manager.set_settings("component_settings.Session.max_length", 3000)
        return self.agent

    def active(self, session_id: str=None):
        if self.current_session_id != None:
            self.stop()
        if session_id == None:
            self.current_session_id = str(uuid.uuid1())
        else:
            self.current_session_id = session_id
            full_chat_history = self.agent.agent_storage.table("chat_history").get(session_id)
            full_chat_history = full_chat_history if full_chat_history else []
            self.full_chat_history.extend(full_chat_history)
            self.recent_chat_history.extend(full_chat_history)
        return self.current_session_id

    def stop(self):
        if self.agent.plugin_manager.get_settings("component_settings.Session.auto_save"):
            self.agent.agent_storage.table("chat_history").set(self.current_session_id, self.full_chat_history).save()
        self.current_session_id = None
        self.full_chat_history = []
        self.recent_chat_history = []
        return self.agent

    def shorten_chat_history(self):
        if len(str(self.recent_chat_history)) > self.agent.plugin_manager.get_settings("component_settings.Session.max_length"):
            self.recent_chat_history = self.recent_chat_history[1:]
            self.shorten_chat_history()
        else:
            return self.agent

    def _prefix(self):
        if self.current_session_id != None:
            self.shorten_chat_history()
            return ("chat_history", self.recent_chat_history)

    def _suffix(self, event: str, data: any):
        if self.current_session_id != None:
            if event == "response:finally":
                new_chat_history = [
                    { "role": "user", "content": data["prompt"]["input"] },
                    { "role": "assistant", "content": data["reply"] }
                ]
                self.full_chat_history.extend(new_chat_history)
                self.recent_chat_history.extend(new_chat_history)

    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": self._suffix,
            "alias": {
                "active_session": { "func": self.active, "return_value": True },
                "stop_session": { "func": self.stop },
                "toggle_session_auto_save": { "func": self.toggle_auto_save },
                "set_chat_history_max_length": { "func": self.set_max_length },
            },
        }

def export():
    return ("Session", Session)


import threading
class ThreadWithResult(threading.Thread):
    def __init__(self, target, args=()):
        super(ThreadWithResult, self).__init__()
        self.target = target
        self.args = args
        self.result = None

    def run(self):
        self.result = self.target(*self.args)


import os
import sys
import importlib

def install_plugins(plugin_manager):
    dir_path = os.path.dirname(os.path.abspath(__file__))
    for item in os.listdir(dir_path):
        if os.path.exists(f"{ dir_path }/{ item }/__init__.py"):
            module_plugins = importlib.import_module(f".{ item }", package = __package__)
            if not hasattr(module_plugins, "export"):
                raise Exception(f"[Plugin Manager] Function 'export' must be stated in '__init__.py' file in module plugin dir: { dir_path }/{ item }")
            module_plugin_export = getattr(module_plugins, "export")()
            module_plugin_list = module_plugin_export[0]
            module_default_settings = module_plugin_export[1]
            try:
                for plugin_info in module_plugin_list:
                    plugin_manager.register(plugin_info[0], plugin_info[1], plugin_info[2])
            except:
                raise Exception(f"[Plugin Manager] Function 'export' in '__init__.py' must return a list of tuple (<module name>, <plugin name>, <plugin method>).")
            if module_default_settings:
                plugin_manager.update_settings(module_default_settings)

def install(Agently):
    install_plugins(Agently.global_plugin_manager)
    Agently.facility.refresh_plugins()

def install_to_agent(agent):
    install_plugins(agent.plugin_manager)
    agent.refresh_plugins()


from .utils import componentABC
from Agently.utils import RuntimeCtxNamespace

class Role(componentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.role_runtime_ctx = RuntimeCtxNamespace("role", self.agent.agent_runtime_ctx)
        self.role_storage = self.agent.global_storage.table("role")
        self.is_enabled = lambda: self.agent.plugin_manager.get_settings("component_toggles.Role")

    def toggle(self, is_enabled: bool):
        self.agent.plugin_manager.set_settings("component_toggles.Role", is_enabled)
        return self.agent

    def set_name(self, name: str):
        self.role_runtime_ctx.set("NAME", name)
        return self.agent

    def set(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.set(key, value)
        else:
            self.role_runtime_ctx.set("ROLE", key)
        return self.agent

    def update(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.update(key, value)
        else:
            self.role_runtime_ctx.update("ROLE", key)
        return self.agent        

    def append(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.append(key, value)
        else:
            self.role_runtime_ctx.append("ROLE", key)
        return self.agent

    def extend(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.extend(key, value)
        else:
            self.role_runtime_ctx.extend("ROLE", key)
        return self.agent

    def save(self, role_name: str=None):
        if role_name == None:
            role_name = self.role_runtime_ctx.get("NAME")
        if role_name != None and role_name != "":
            self.role_storage\
                .set(role_name, self.role_runtime_ctx.get())\
                .save()
            return self.agent
        else:
            raise Exception("[Agent Component: Role] Role attr 'NAME' must be stated before save.")

    def load(self, role_name: str):
        role_data = self.role_storage.get(role_name)
        self.role_runtime_ctx.update(role_data)
        return self.agent

    def _prefix(self):
        if not self.is_enabled():
            return None
        role_data = self.role_runtime_ctx.get()
        if role_data != None:
            return ("system", role_data)
        else:
            return None

    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": None,
            "alias": {
                "toggle_role": { "func": self.toggle },
                "set_role_name": { "func": self.set_name },
                "set_role": { "func": self.set },
                "update_role": { "func": self.update },
                "append_role": { "func": self.append },
                "extend_role": { "func": self.extend },
                "save_role": { "func": self.save },
                "load_role": { "func": self.load },
            },
        }

def export():
    return ("Role", Role)


from .utils import PluginManager, StorageDelegate
from .plugins import install_plugins
from .WebSocket import WebSocketServer

global_plugin_manager = PluginManager()
install_plugins(global_plugin_manager)
global_storage = StorageDelegate(
    db_name = "global",
    plugin_manager = global_plugin_manager,
)
global_websocket_server = WebSocketServer()


from .Request import Request
from .Agent import AgentFactory
from .Facility import FacilityManager
from .WebSocket import WebSocketServer, WebSocketClient
from ._global import global_plugin_manager, global_storage
from .utils import *

def create_agent(*args, **kwargs):
    return AgentFactory().create_agent(*args, **kwargs)

facility = FacilityManager()


from Agently.utils import RuntimeCtx

class RoleManager(object):
    def __init__(self, *, storage: object, plugin_manager: object):
        self.storage = storage.table("role")
        self.plugin_manager = plugin_manager
        self.role_runtime_ctx = RuntimeCtx()

    def name(self, name: str):
        self.role_runtime_ctx.set("NAME", name)
        return self

    def set(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.set(key, value)
        else:
            self.role_runtime_ctx.set("ROLE", key)
        return self

    def update(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.update(key, value)
        else:
            self.role_runtime_ctx.update("ROLE", key)
        return self     

    def append(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.append(key, value)
        else:
            self.role_runtime_ctx.append("ROLE", key)
        return self

    def extend(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.extend(key, value)
        else:
            self.role_runtime_ctx.extend("ROLE", key)
        return self

    def save(self, role_name: str=None):
        if role_name == None:
            role_name = self.role_runtime_ctx.get("NAME")
        if role_name != None and role_name != "":
            self.storage\
                .set(role_name, self.role_runtime_ctx.get())\
                .save()
            self.role_runtime_ctx.empty()
            return self
        else:
            raise Exception("[Facility: RoleMananger] Role attr 'NAME' must be stated before save.")

    def get(self, role_name: str):
        return self.storage.get(role_name)

def export():
    return ("role_manager", RoleManager)


import json
import inspect

class AliasManager(object):
    def __init__(self, target: object):
        self.target = target
        self.alias_name_list = []
        self.alias_details = {}

    def register(self, alias_name: str, alias_func: callable, *, return_value: bool=False):
        if alias_name in self.alias_name_list:
            raise Exception(f"[Alias Manager] alias_name '{ alias_name }' has already been registered.")
        if return_value:
            setattr(self.target, alias_name, alias_func)
        else:
            def _alias_func(*args, **kwargs):
                alias_func(*args, **kwargs)
                return self.target
            setattr(self.target, alias_name, _alias_func)
        self.alias_name_list.append(alias_name)
        self.alias_details.update({ alias_name: { "func": alias_func, "return_value": return_value } })

    def get_alias_info(self):
        result = {}
        for alias_name, alias_info in self.alias_details.items():
            result[alias_name] = { "paramaters": [] }
            alias_func_info = inspect.signature(alias_info["func"])
            for name, meta in alias_func_info.parameters.items():
                result[alias_name]["paramaters"].append(str(meta.name))
            result[alias_name]["is_return_value"] = alias_info["return_value"]
            result[alias_name]["docstring"] = str(alias_info["func"].__doc__)
        return result

    def print_alias_info(self):
        print(json.dumps(self.get_alias_info(), indent=4, ensure_ascii=False))


from .RuntimeCtx import RuntimeCtx, RuntimeCtxNamespace
from .StorageDelegate import StorageDelegate
from .PluginManager import PluginManager
from .AliasManager import AliasManager
from .IdGenerator import IdGenerator
from .DataOps import DataOps, NamespaceOps
from .ThreadWithResult import ThreadWithResult
from .transform import to_json_desc, to_instruction, find_all_jsons, find_json
from .check_version import check_version


import yaml

def to_instruction(origin):
    if origin == None:
        return None
    elif isinstance(origin, (list, tuple, set, dict)):
        return yaml.dump(origin, allow_unicode=True, sort_keys=False)
    else:
        return str(origin)

def to_json_desc(origin, layer_count = 0):
    if isinstance(origin, dict):
        json_string = ""
        if layer_count > 0:
            json_string += "\n"
        json_string += ("\t" * layer_count) + "{\n"
        for key, value in origin.items():
            json_string += ("\t" * (layer_count + 1)) + "\"" + key + "\": " + to_json_desc(value, layer_count + 1) + "\n"
        if layer_count > 0:
            json_string += ("\t" * (layer_count + 1)) + "},"
        else:
            json_string += "}"
        return json_string
    elif isinstance(origin, (list, set)):
        json_string = ""
        if layer_count > 0:
            json_string += "\n"
        json_string += ("\t" * layer_count) + "[\n"
        for item in origin:
            json_string += ("\t" * (layer_count + 1)) + to_json_desc(item, layer_count + 1) + ",\n"
        json_string += ("\t" * (layer_count + 1)) + "...\n"
        if layer_count > 0:
            json_string += ("\t" * layer_count) + "],"
        else:
            json_string += "]"
        return json_string
    elif isinstance(origin, tuple):
        json_string = f"<{ origin[0] }>,"
        if len(origin) >= 2:
            json_string += f"//{ origin[1] }"
        return json_string
    else:
        return str(origin)

def find_all_jsons(origin: str):
    stage = 1
    json_blocks = []
    block_num = 0
    layer = 0
    skip_next = False
    in_quote = False
    for char in origin:
        if skip_next:
            skip_next = False
            continue
        if stage == 1:
            if char == "\\":
                skip_next = True
                continue
            if char == "[" or char == "{":
                json_blocks.append(char)
                stage = 2
                layer += 1
                continue
        elif stage == 2:
            if char == "\\":
                skip_next = True
                continue
            if char == "\"":
                in_quote = not in_quote
            if char == "[" or char == "{":
                json_blocks[block_num] += char
                layer += 1
            elif char == "]" or char == "}":
                json_blocks[block_num] += char
                layer -= 1
            else:
                if in_quote:
                    json_blocks[block_num] += char
                elif char not in (" ", "\t", "\n"):
                    json_blocks[block_num] += char
            if layer == 0:
                block_num += 1
                stage = 1
    return json_blocks

def find_json(origin: str):
    result = find_all_jsons(origin)
    if len(result) > 0:
        return result[0]
    else:
        return None


from .transform import to_instruction, to_json_desc

def construct_prompt_from_dict(prompt_dict, layer_count = 1):
    prompt = ""
    for key, content in prompt_dict.items():
        prompt += f"{ '#' * layer_count } { key }:\n"
        if isinstance(content, dict):
            prompt += construct_prompt_from_dict(content, layer_count + 1) + '\n'
        else:
            prompt += str(content) + '\n'
        if layer_count == 1:
            prompt += "\n"
    if layer_count == 1:
        prompt += "# [OUTPUT]:\n"
    return prompt

def construct_request_messages(request_runtime_ctx):
    #init request messages
    request_messages = []
    # - system message
    system_data = request_runtime_ctx.get("system")
    if system_data:
        request_messages.append({ "role": "system", "content": to_instruction(system_data) })
    # - headline
    headline_data = request_runtime_ctx.get("headline")
    if headline_data:
        request_messages.append({ "role": "assistant", "content": to_instruction(headline_data) })
    # - chat history
    chat_history_data = request_runtime_ctx.get("chat_history")
    if chat_history_data:
        request_messages.extend(chat_history_data)
    # - request message (prompt)
    prompt_input_data = request_runtime_ctx.get("prompt_input")
    prompt_information_data = request_runtime_ctx.get("prompt_information")
    prompt_instruction_data = request_runtime_ctx.get("prompt_instruction")
    prompt_output_data = request_runtime_ctx.get("prompt_output")
    # --- only input
    if not prompt_input_data and not prompt_information_data and not prompt_instruction_data and not prompt_output_data:
        raise Exception("[Request] Missing 'prompt_input', 'prompt_information', 'prompt_instruction', 'prompt_output' in request runtime_ctx. At least set value to one of them.")
    if prompt_input_data and not prompt_information_data and not prompt_instruction_data and not prompt_output_data:
        request_messages.append({ "role": "user", "content": to_instruction(prompt_input_data) })
    # --- construct prompt
    else:
        prompt_dict = {}
        if prompt_input_data:
            prompt_dict["[INPUT]"] = to_instruction(prompt_input_data)
        if prompt_information_data:
            prompt_dict["[HELPFUL INFORMATION]"] = to_instruction(prompt_information_data)
        if prompt_instruction_data:
            prompt_dict["[INSTRUCTION]"] = to_instruction(prompt_instruction_data)
        if prompt_output_data:
            if isinstance(prompt_output_data, (dict, list, set)):
                prompt_dict["[OUTPUT REQUIREMENT]"] = {
                    "TYPE": "JSON can be parsed in Python",
                    "FORMAT": to_json_desc(prompt_output_data),
                }
                request_runtime_ctx.set("response:type", "JSON")
            else:
                prompt_dict["[OUTPUT REQUIERMENT]"] = str(prompt_output_data)
        request_messages.append({ "role": "user", "content": construct_prompt_from_dict(prompt_dict) })
    return request_messages


import yaml

def to_instruction(origin):
    if origin == None:
        return None
    elif isinstance(origin, (list, tuple, set, dict)):
        return yaml.dump(origin, allow_unicode=True, sort_keys=False)
    else:
        return str(origin)

def to_json_desc(origin, layer_count = 0):
    if isinstance(origin, dict):
        json_string = ""
        if layer_count > 0:
            json_string += "\n"
        json_string += ("\t" * layer_count) + "{\n"
        for key, value in origin.items():
            json_string += ("\t" * (layer_count + 1)) + "\"" + key + "\": " + to_json_desc(value, layer_count + 1) + "\n"
        if layer_count > 0:
            json_string += ("\t" * (layer_count + 1)) + "},"
        else:
            json_string += "}"
        return json_string
    elif isinstance(origin, (list, set)):
        json_string = ""
        if layer_count > 0:
            json_string += "\n"
        json_string += ("\t" * layer_count) + "[\n"
        for item in origin:
            json_string += ("\t" * (layer_count + 1)) + to_json_desc(item, layer_count + 1) + ",\n"
        json_string += ("\t" * (layer_count + 1)) + "...\n"
        if layer_count > 0:
            json_string += ("\t" * layer_count) + "],"
        else:
            json_string += "]"
        return json_string
    elif isinstance(origin, tuple):
        json_string = f"<{ origin[0] }>,"
        if len(origin) >= 2:
            json_string += f"//{ origin[1] }"
        return json_string
    else:
        return str(origin)

def find_json(origin: str):
    stack = []
    result = ""
    found_open_bracket = False

    for char in origin:
        if char not in ("\t","\n"):
            if found_open_bracket:
                result += char

            if char == '{':
                found_open_bracket = True
                result += char
                stack.append(char)
            elif char == '}':
                if found_open_bracket:
                    stack.pop()
                    if not stack:
                        break

    if len(result) >= 2 and result[-2] == ",":
        result = result[:-2] + "}"
    return result if stack == [] else None